@@ -8,6 +8,7 @@ import netsquid as ns
 
 from qoala.lang.iqoala import IqoalaParser, IqoalaProgram
 from qoala.runtime.config import (
+    DepolariseLinkConfig,
     GenericQDeviceConfig,
     LinkConfig,
     ProcNodeConfig,
@@ -24,6 +25,10 @@ from qoala.runtime.schedule import (
 from qoala.sim.build import build_network
 from qoala.sim.network import ProcNodeNetwork
 
+INSTR_LATENCY = 1e5
+CC_LATENCY = 1e6
+QC_EXPECTATION = 30e6
+
 
 def create_global_env(
     num_clients: int, global_schedule: List[int], timeslot_len: int
@@ -46,17 +51,23 @@ def get_client_config(id: int) -> ProcNodeConfig:
         node_id=id,
         qdevice_typ="generic",
         qdevice_cfg=GenericQDeviceConfig.perfect_config(1),
-        instr_latency=1000,
+        instr_latency=INSTR_LATENCY,
+        receive_latency=CC_LATENCY,
     )
 
 
 def get_server_config(id: int, num_qubits: int) -> ProcNodeConfig:
+    config_file = relative_to_cwd("node_config.yaml")
+    qdevice_cfg = GenericQDeviceConfig.from_file(config_file)
+    qdevice_cfg.num_qubits = num_qubits
     return ProcNodeConfig(
         name="server",
         node_id=id,
         qdevice_typ="generic",
-        qdevice_cfg=GenericQDeviceConfig.perfect_config(num_qubits),
-        instr_latency=1000,
+        # qdevice_cfg=GenericQDeviceConfig.perfect_config(num_qubits),
+        qdevice_cfg=qdevice_cfg,
+        instr_latency=INSTR_LATENCY,
+        receive_latency=CC_LATENCY,
     )
 
 
@@ -71,8 +82,13 @@ def create_network(
 
     global_env = create_global_env(num_clients, global_schedule, timeslot_len)
 
+    link_cfg_file = relative_to_cwd("link_config.yaml")
+    depolarise_config = DepolariseLinkConfig.from_file(link_cfg_file)
     link_cfgs = [
-        LinkConfig.perfect_config("server", cfg.name) for cfg in client_configs
+        LinkConfig(
+            node1="server", node2=cfg.name, typ="depolarise", cfg=depolarise_config
+        )
+        for cfg in client_configs
     ]
 
     node_cfgs = [server_cfg] + client_configs
@@ -84,11 +100,10 @@ def create_network(
 @dataclass
 class TaskDurations:
     instr_latency: int
-    rot_dur: int
-    h_dur: int
-    meas_dur: int
-    free_dur: int
-    cphase_dur: int
+    cc_latency: int
+    single_gate: int
+    two_gate: int
+    meas: int
 
 
 def create_server_tasks(
@@ -96,17 +111,16 @@ def create_server_tasks(
 ) -> ProgramTaskList:
     tasks = []
 
-    cl_dur = 1e3
-    cc_dur = 10e6
-    # ql_dur = 1e4
-    qc_dur = 1e6
+    cl_dur = 2 * task_durations.instr_latency
+    cc_dur = 2 * task_durations.cc_latency
+    qc_dur = QC_EXPECTATION
 
     set_dur = task_durations.instr_latency
-    rot_dur = task_durations.rot_dur
-    h_dur = task_durations.h_dur
-    meas_dur = task_durations.meas_dur
-    free_dur = task_durations.free_dur
-    cphase_dur = task_durations.cphase_dur
+    rot_dur = task_durations.single_gate
+    h_dur = task_durations.single_gate
+    meas_dur = task_durations.meas
+    free_dur = task_durations.instr_latency
+    cphase_dur = task_durations.two_gate
 
     # csocket = assign_cval() : 0
     tasks.append(TaskBuilder.CL(cl_dur, 0))
@@ -141,10 +155,9 @@ def create_server_tasks(
     tasks.append(TaskBuilder.QL(h_dur, "meas_qubit_0", 2))
     tasks.append(TaskBuilder.QL(meas_dur, "meas_qubit_0", 3))
     tasks.append(TaskBuilder.QL(free_dur, "meas_qubit_0", 4))
-    # return_result(m1)
-    tasks.append(TaskBuilder.CL(cl_dur, 9))
-    # return_result(m2)
-    tasks.append(TaskBuilder.CL(cl_dur, 10))
+
+    # send_cmsg(csocket, m2)
+    tasks.append(TaskBuilder.CC(cc_dur, 9))
 
     return ProgramTaskList(server_program, {i: task for i, task in enumerate(tasks)})
 
@@ -154,76 +167,132 @@ def create_client_tasks(
 ) -> ProgramTaskList:
     tasks = []
 
-    cl_dur = 1e3
-    cc_dur = 10e6
-    # ql_dur = 1e3
-    qc_dur = 1e6
+    cl_dur = task_durations.instr_latency
+    cc_dur = task_durations.cc_latency
+    qc_dur = QC_EXPECTATION
 
     set_dur = task_durations.instr_latency
-    rot_dur = task_durations.rot_dur
-    h_dur = task_durations.h_dur
-    meas_dur = task_durations.meas_dur
-    free_dur = task_durations.free_dur
+    rot_dur = task_durations.single_gate
+    h_dur = task_durations.single_gate
+    meas_dur = task_durations.meas
+    free_dur = task_durations.instr_latency
 
-    tasks.append(TaskBuilder.CL(cl_dur, 0))
-    tasks.append(TaskBuilder.CL(cl_dur, 1))
+    class Counter:
+        def __init__(self):
+            self.index = 0
+
+        def next(self):
+            index = self.index
+            self.index += 1
+            return index
+
+    c = Counter()
+
+    # csocket = assign_cval() : 0
+    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
+    # const_1 = assign_cval() : 1
+    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
+
+    # compute epr0_rot_y etc
+    for _ in range(10):
+        tasks.append(TaskBuilder.CL(cl_dur, c.next()))
+
+    # run_subroutine(vec<>) : create_epr_0
+    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
+    # vec<p2> = run_subroutine(vec<theta2>) : post_epr_0
     tasks.append(TaskBuilder.QC(qc_dur, "create_epr_0"))
-    tasks.append(TaskBuilder.CL(cl_dur, 2))
+
+    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
     tasks.append(TaskBuilder.QL(set_dur, "post_epr_0", 0))
     tasks.append(TaskBuilder.QL(rot_dur, "post_epr_0", 1))
     tasks.append(TaskBuilder.QL(h_dur, "post_epr_0", 2))
     tasks.append(TaskBuilder.QL(meas_dur, "post_epr_0", 3))
     tasks.append(TaskBuilder.QL(free_dur, "post_epr_0", 4))
+    tasks.append(TaskBuilder.QL(free_dur, "post_epr_0", 5))
 
-    tasks.append(TaskBuilder.CL(cl_dur, 3))
+    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
     tasks.append(TaskBuilder.QC(qc_dur, "create_epr_1"))
-    tasks.append(TaskBuilder.CL(cl_dur, 4))
+    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
     tasks.append(TaskBuilder.QL(set_dur, "post_epr_1", 0))
     tasks.append(TaskBuilder.QL(rot_dur, "post_epr_1", 1))
     tasks.append(TaskBuilder.QL(h_dur, "post_epr_1", 2))
     tasks.append(TaskBuilder.QL(meas_dur, "post_epr_1", 3))
     tasks.append(TaskBuilder.QL(free_dur, "post_epr_1", 4))
+    tasks.append(TaskBuilder.QL(free_dur, "post_epr_1", 5))
 
-    tasks.append(TaskBuilder.CL(cl_dur, 5))
-    tasks.append(TaskBuilder.CL(cl_dur, 6))
-    tasks.append(TaskBuilder.CL(cl_dur, 7))
-    tasks.append(TaskBuilder.CL(cl_dur, 8))
-    tasks.append(TaskBuilder.CC(cc_dur, 9))
-    tasks.append(TaskBuilder.CC(cc_dur, 10))
-    tasks.append(TaskBuilder.CL(cl_dur, 11))
-    tasks.append(TaskBuilder.CL(cl_dur, 12))
-    tasks.append(TaskBuilder.CL(cl_dur, 13))
-    tasks.append(TaskBuilder.CL(cl_dur, 14))
-    tasks.append(TaskBuilder.CL(cl_dur, 15))
-    tasks.append(TaskBuilder.CC(cc_dur, 16))
-    tasks.append(TaskBuilder.CL(cl_dur, 17))
-    tasks.append(TaskBuilder.CL(cl_dur, 18))
+    # x = mult_const(p1) : 16
+    # minus_theta1 = mult_const(theta1) : -1
+    # delta1 = add_cval_c(minus_theta1, x)
+    # delta1 = add_cval_c(delta1, alpha)
+    for _ in range(4):
+        tasks.append(TaskBuilder.CL(cl_dur, c.next()))
+
+    # minus_dummy0 = mult_const(dummy0) : -1
+    # should_correct_0 = add_cval_c(const_1, minus_dummy0)
+    # delta1_correction = bcond_mult_const(alpha, should_correct_0) : 0
+    # delta1_correction = mult_const(delta1_correction) : -1
+    # delta1 = add_cval_c(delta1, delta1_correction)
+    for _ in range(5):
+        tasks.append(TaskBuilder.CL(cl_dur, c.next()))
+
+    # send_cmsg(csocket, delta1)
+    # m1 = recv_cmsg(csocket)
+    tasks.append(TaskBuilder.CC(cl_dur, c.next()))
+    tasks.append(TaskBuilder.CC(cc_dur, c.next()))
+
+    # y = mult_const(p2) : 16
+    # minus_theta2 = mult_const(theta2) : -1
+    # beta = bcond_mult_const(beta, m1) : -1
+    # delta2 = add_cval_c(beta, minus_theta2)
+    # delta2 = add_cval_c(delta2, y)
+    for _ in range(5):
+        tasks.append(TaskBuilder.CL(cl_dur, c.next()))
+
+    # minus_dummy1 = mult_const(dummy1) : -1
+    # should_correct_1 = add_cval_c(const_1, minus_dummy1)
+    # delta2_correction = bcond_mult_const(beta, should_correct_1) : 0
+    # delta2_correction = mult_const(delta2_correction) : -1
+    # delta2 = add_cval_c(delta2, delta2_correction)
+    for _ in range(5):
+        tasks.append(TaskBuilder.CL(cl_dur, c.next()))
+
+    # send_cmsg(csocket, delta2)
+    # m2 = recv_cmsg(csocket)
+    tasks.append(TaskBuilder.CC(cl_dur, c.next()))
+    tasks.append(TaskBuilder.CC(cl_dur, c.next()))
+
+    # return results
+    for _ in range(4):
+        tasks.append(TaskBuilder.CL(cl_dur, c.next()))
 
     return ProgramTaskList(client_program, {i: task for i, task in enumerate(tasks)})
 
 
 @dataclass
 class BqcResult:
-    server_batches: Dict[int, ProgramBatch]
-    server_results: Dict[int, BatchResult]
+    client_batches: List[Dict[int, ProgramBatch]]
+    client_results: List[Dict[int, BatchResult]]
 
 
 def create_durations() -> TaskDurations:
-    perfect_qdevice_cfg = GenericQDeviceConfig.perfect_config(1)
-    instr_latency = 1000
+    config_file = relative_to_cwd("node_config.yaml")
+    qdevice_cfg = GenericQDeviceConfig.from_file(config_file)
 
     return TaskDurations(
-        instr_latency=instr_latency,
-        rot_dur=perfect_qdevice_cfg.single_qubit_gate_time,
-        h_dur=perfect_qdevice_cfg.single_qubit_gate_time,
-        meas_dur=perfect_qdevice_cfg.measure_time,
-        free_dur=instr_latency,
-        cphase_dur=perfect_qdevice_cfg.two_qubit_gate_time,
+        instr_latency=INSTR_LATENCY,
+        cc_latency=CC_LATENCY,
+        single_gate=qdevice_cfg.single_qubit_gate_time,
+        two_gate=qdevice_cfg.two_qubit_gate_time,
+        meas=qdevice_cfg.measure_time,
     )
 
 
+def relative_to_cwd(file: str) -> str:
+    return os.path.join(os.path.dirname(__file__), file)
+
+
 def load_server_program(remote_name: str) -> IqoalaProgram:
-    path = os.path.join(os.path.dirname(__file__), "bqc_server.iqoala")
+    path = relative_to_cwd("bqc_server.iqoala")
     with open(path) as file:
         server_text = file.read()
     program = IqoalaParser(server_text).parse()
@@ -236,7 +305,7 @@ def load_server_program(remote_name: str) -> IqoalaProgram:
 
 
 def load_client_program() -> IqoalaProgram:
-    path = os.path.join(os.path.dirname(__file__), "bqc_client.iqoala")
+    path = relative_to_cwd("bqc_client.iqoala")
     with open(path) as file:
         client_text = file.read()
     return IqoalaParser(client_text).parse()
@@ -283,6 +352,8 @@ def run_bqc(
     beta,
     theta1,
     theta2,
+    dummy0,
+    dummy1,
     num_iterations: List[int],
     deadlines: List[int],
     num_clients: int,
@@ -332,6 +403,8 @@ def run_bqc(
                     "beta": beta,
                     "theta1": theta1,
                     "theta2": theta2,
+                    "dummy0": dummy0,
+                    "dummy1": dummy1,
                 }
             )
             for _ in range(num_iterations[index])
@@ -352,12 +425,13 @@ def run_bqc(
     end_time = ns.sim_time()
     makespan = end_time - start_time
 
-    server_batches = server_procnode.get_batches()
+    client_procnodes = [network.nodes[f"client_{i}"] for i in range(1, num_clients + 1)]
+    client_batches = [node.get_batches() for node in client_procnodes]
 
-    server_results: Dict[int, BatchResult]
-    server_results = server_procnode.scheduler.get_batch_results()
+    client_results: List[Dict[int, BatchResult]]
+    client_results = [node.scheduler.get_batch_results() for node in client_procnodes]
 
-    return BqcResult(server_batches, server_results), makespan
+    return BqcResult(client_batches, client_results), makespan
 
 
 def check(
@@ -365,6 +439,8 @@ def check(
     beta,
     theta1,
     theta2,
+    dummy0,
+    dummy1,
     expected,
     num_iterations,
     deadlines,
@@ -378,6 +454,8 @@ def check(
         beta=beta,
         theta1=theta1,
         theta2=theta2,
+        dummy0=dummy0,
+        dummy1=dummy1,
         num_iterations=num_iterations,
         deadlines=deadlines,
         num_clients=num_clients,
@@ -387,15 +465,25 @@ def check(
 
     batch_success_probabilities: List[float] = []
 
-    batch_results = bqc_result.server_results
-    for batch_id, batch_result in batch_results.items():
-        program_batch = bqc_result.server_batches[batch_id]
+    for i in range(num_clients):
+        assert len(bqc_result.client_results[i]) == 1
+        batch_result = bqc_result.client_results[i][0]
+        assert len(bqc_result.client_batches[i]) == 1
+        program_batch = bqc_result.client_batches[i][0]
         batch_iterations = program_batch.info.num_iterations
-        program_results = batch_result.results
-        m2s = [result.values["m2"] for result in program_results]
-        correct_outcomes = len([m2 for m2 in m2s if m2 == expected])
-        succ_prob = round(correct_outcomes / batch_iterations, 2)
-        batch_success_probabilities.append(succ_prob)
+
+        p1s = [result.values["p1"] for result in batch_result.results]
+        p2s = [result.values["p2"] for result in batch_result.results]
+        m1s = [result.values["m1"] for result in batch_result.results]
+        m2s = [result.values["m2"] for result in batch_result.results]
+
+        if dummy0 == 0:
+            num_fails = len([(p, m) for (p, m) in zip(p1s, m2s) if p != m])
+        else:  # dummy0 = 1
+            num_fails = len([(p, m) for (p, m) in zip(p2s, m1s) if p != m])
+
+        frac_fail = round(num_fails / batch_iterations, 2)
+        batch_success_probabilities.append(1 - frac_fail)
 
     return batch_success_probabilities, makespan
 
@@ -414,6 +502,8 @@ def compute_succ_prob(
         beta=24,
         theta1=2,
         theta2=22,
+        dummy0=0,
+        dummy1=1,
         expected=1,
         num_iterations=num_iterations,
         deadlines=deadlines,
@@ -424,17 +514,13 @@ def compute_succ_prob(
 
 
 def test_bqc():
-
-    # 5 clients, i.e. 5 BQC client applications (one on each client node)
-    # and 5 BQC server applications (all 5 on the single server node)
-    # Do 10 iterations for each of the 5 BQC applications.
-    # All applications have a deadline of 1e9.
+    num_clients = 10
     succ_probs, makespan = compute_succ_prob(
-        num_clients=5,
-        num_iterations=[10, 10, 10, 10, 10],
-        deadlines=[1e9, 1e9, 1e9, 1e9, 1e9],
-        global_schedule=[1, 2, 3, 4, 5],
-        timeslot_len=1e6,
+        num_clients=num_clients,
+        num_iterations=[30] * num_clients,
+        deadlines=[1e8] * num_clients,
+        global_schedule=[i for i in range(num_clients)],
+        timeslot_len=50e6,
     )
     print(f"success probabilities: {succ_probs}")
     print(f"makespan: {makespan}")
